#!/usr/bin/php
<?php
if (php_sapi_name() !== 'cli')
	exit('Must be run from CLI.' . PHP_EOL);

const LF = "\n";

define('ROOT', dirname($_SERVER['SCRIPT_FILENAME']));

if (!extension_loaded('tidy'))
	echo 'PHP tidy extension unavailable. Feature disabled.' . PHP_EOL;

foreach (['pandoc', 'gzip'] as $command) {
	exec('command -v ' . $command, result_code: $code);
	if ($code !== 0)
		exit($command . ' command not available.' . PHP_EOL);
}

foreach ($argv as $arg) {
	if ($arg === '-f')
		$opt['force'] = true;
	else
		$args[] = $arg;
}
$opt['force'] ??= false;

define('SITE', $args[1] ?? getcwd());

define('DESTINATION', $args[2] ?? 'dns');

if (file_exists(SITE . '/config.ini'))
	$config = parse_ini_file(SITE . '/config.ini');

$config['header'] ??= false;
$config['author'] ??= NULL;
$config['base-url'] ??= [];
$config['center-index'] ??= false;
$config['default-lang'] ??= NULL;
$config['announce-css'] ??= false;
$config['announce-feed'] ??= false;

if (!isset($config['id'])) {
	$config['id'] = bin2hex(random_bytes(32));
	file_put_contents(SITE . '/config.ini', 'id = "' . $config['id'] . '"' . LF, FILE_APPEND);
}

if ($config['announce-css'])
	copy(ROOT . '/style.css', SITE . '/mkht-php.css');

// Determine whether links need to use Onion or DNS
function clearnetOrOnion($clearnet_url, $onion_url) {
	return (DESTINATION === 'onion') ? $onion_url : $clearnet_url;
}

$files = new RecursiveIteratorIterator(new RecursiveDirectoryIterator(SITE . '/src', RecursiveDirectoryIterator::SKIP_DOTS));

foreach($files as $file) {
	$info = new SplFileInfo($file->getPathName());
	if ($info->getType() !== 'file' OR !in_array($info->getExtension(), ['gmi', 'md', 'html'], true) OR str_starts_with($info->getPathname(), '.'))
		continue;
	$files_dates[$info->getPathname()] = $info->getMTime();
}

asort($files_dates);

ob_start();
?>
<?xml version="1.0" encoding="UTF-8"?>
<feed xmlns="http://www.w3.org/2005/Atom">
	<title><?= $config['title'] ?? '' ?></title>
	<id>urn:publicid:<?= $config['id'] ?></id>
<?php
foreach ($config['base-url'] as $url)
	echo '	<link rel="self" type="application/atom+xml" href="' . $url . '/feed.atom"></link>' . LF;
?>
	<updated><?= date('c', $files_dates[array_key_last($files_dates)]) ?></updated>
	<author>
		<name><?= $config['author'] ?? '' ?></name>
	</author>
<?php
$feed = ob_get_clean();

foreach ($files_dates as $src_page => $last_mod) {
	$content = file_get_contents($src_page);

	preg_match('/^# ?(?<title>.*)$/Dm', $content, $matches);
	$title = $matches['title'] ?? NULL;

	$path_parts = pathinfo(str_replace('/src/', '/', $src_page));

	$base_filepath = $path_parts['dirname'] . '/' . $path_parts['filename'];

	if (!file_exists($base_filepath . '.html') OR (filemtime($src_page) > filemtime($base_filepath . '.html')) OR $opt['force']) {
		echo 'Compiling ' . $src_page . ' ' . date("Y-m-d H:i:s", $last_mod) . LF;

		// Create parent directory if needed
		if (!file_exists($path_parts['dirname']))
			mkdir($path_parts['dirname'], 0755, true);

		// Execute PHP code
		ob_start();
		eval('?>' . $content);
		$content = ob_get_clean();
		file_put_contents($base_filepath . '.' . $path_parts['extension'], $content);

		// Convert Gemtext to Markdown
		if ($path_parts['extension'] === 'gmi') {
			$content = preg_replace_callback(
				'/^=>\h*(?<addr>\S+)(:?\h+(?<title>\V+))?$/m',
				function ($matches) {
					if (!str_contains($matches['addr'], ':') AND str_ends_with($matches['addr'], '.gmi'))
						$matches['addr'] = substr($matches['addr'], 0, -3) . 'md';
					return '[' . ($matches['title'] ?? $matches['addr']) . '](' . $matches['addr'] . ')';
				},
				$content,
			);
			file_put_contents($base_filepath . '.md', $content);
		}

		// Compile Markdown to HTML
		$process = proc_open('pandoc --fail-if-warnings -f markdown_phpextra-citations-native_divs-native_spans+abbreviations+hard_line_breaks+lists_without_preceding_blankline -t html --wrap none', [
			0 => ['pipe', 'r'],
			1 => ['pipe', 'w'],
		], $pipes);
		if (is_resource($process) !== true)
			exit('Can\'t spawn pandoc.' . PHP_EOL);
		fwrite($pipes[0], $content);
		fclose($pipes[0]);
		$content = stream_get_contents($pipes[1]);
		fclose($pipes[1]);
		if (proc_close($process) !== 0)
			exit('pandoc failed.' . PHP_EOL);

		// .md > .html for local links
		$content = preg_replace('/ href="([^:"]+)\.md"/', ' href="$1.html"', $content);

		$relative_root_path = str_repeat('../', substr_count(str_replace(SITE, '', $path_parts['dirname']), '/'));

		ob_start();

	?>
		<!DOCTYPE html>
		<html lang="<?php

		preg_match('#\.([a-zA-Z-]{2,5})\.#', $path_parts['basename'], $file_lang);
		if (isset($file_lang[1])) {
			$lang = $file_lang[1];
		} else {
			preg_match('#/([a-z]{2})(/|$)#', $path_parts['dirname'], $dir_lang);
			$lang = $dir_lang[1] ?? $config['default-lang'];
		}
		echo $lang ?? '';

		?>">
			<head>
				<meta charset="utf-8">
				<?php
		if (isset($title) AND isset($config['title']))
			echo '<title>' . $title . ' · ' . $config['title'] . '</title>';
		else if (isset($title))
			echo '<title>' . $title . '</title>';
		else if (isset($config['title']))
			echo '<title>' . $config['title'] . '</title>';
				?>
				<meta name="viewport" content="width=device-width, initial-scale=1">
				<meta name="referrer" content="no-referrer">
<?php
		if (isset($config['author']))
			echo '<meta name="author" content="' . $config['author'] . '">';

		if ($config['announce-feed'])
			echo '<link rel="alternate" type="application/atom+xml" href="' . $relative_root_path . 'feed.atom">' . LF;

		if ($config['announce-css']) {
			if (file_exists(SITE . '/style.css'))
				echo '<link rel="stylesheet" media="screen" href="' . $relative_root_path . 'style.css">' . LF;
			echo '<link rel="stylesheet" media="screen" href="' . $relative_root_path . 'mkht-php.css">' . LF;
		}

		if (file_exists(SITE . '/head.inc.html'))
			echo file_get_contents(SITE . '/head.inc.html');
	?>
			</head>

			<body>
	<?php
		if ($config['header']) {
	?>
				<header>
					<a href="./<?= $relative_root_path ?>">
	<?php
			if (file_exists(SITE . '/img/logo.webp'))
				echo '<img src="img/logo.webp" ' . getimagesize(SITE . '/img/logo.webp')[3] . ' alt="' . $config['title'] . '" />';
			else
				echo $config['site-title'];
	?>
					</a>
				</header>
	<?php
		}

		if ($config['center-index'] AND $path_parts['filename'] === 'index')
			echo '<div class="centered">' . $content . '</div>';
		else
			echo '<main>' . $content . '</main>';
		if (file_exists(SITE . '/end.inc.html'))
			require SITE . '/end.inc.html';
		echo '</body></html>';

		$content = ob_get_clean();

		if (extension_loaded('tidy')) {
			$content = tidy_repair_string($content, [
				'indent' => true,
				'indent-spaces' => 4,
				'output-xhtml' => true,
				'wrap' => 0,
			]);
			$content = str_replace('    ', '	', $content);
		}

		file_put_contents($base_filepath . '.html', $content);

		// Gzip compression
		exec('gzip --keep --fast --force ' . $base_filepath . '.html');
	}

	$relative_addr = substr_replace($base_filepath . '.html', '', strpos($base_filepath, SITE), strlen(SITE));

	// As of RFC 3151: A URN Namespace for Public Identifiers
	$public_id = 'urn:publicid:' . $config['id'] . str_replace('/', '%2F', $relative_addr);

	preg_match('#\<body\>(?<content>.*)\</body\>#s', file_get_contents($base_filepath . '.html'), $match);
	$atom_entry_content = $match['content'];

	// Make relative links absolute
	$atom_entry_content = preg_replace_callback('# href=\"(?<relative_url>[^:"]+)\"#', function ($matches) {
		global $config;
		global $path_parts;

		return ' href="' . ($config['base-url'][0] ?? '') . substr($path_parts['dirname'], strlen(SITE)) . '/' . $matches['relative_url'] . '"';
	}, $atom_entry_content);

	if (!in_array('draft', explode('.', $path_parts['basename']), true)) {
		ob_start();
?>
	<entry>
		<title><?= $title ?></title>
		<id><?= $public_id ?></id>
		<updated><?= date('c', $last_mod) ?></updated>
<?php
		foreach ($config['base-url'] as $base_url)
			echo '		<link rel="alternate" type="text/html" href="' . $base_url . $relative_addr . '"></link>' . LF;
?>
		<content type="html"><?= htmlspecialchars($atom_entry_content) ?></content>
	</entry>
<?php
		$feed .= ob_get_clean();
	}
}
file_put_contents(SITE . '/feed.atom', $feed . '</feed>' . LF);
